#include "defo.h"

#include "find-symmetry.h"
#include "my-interval-func.h"
#include "bodies.h"
#include "solutions-tests.h"
#include "test-and-help-functions.h"
#include "krawczyk-newton.h"
#include "search-alg.h"

#include "cc-nb.h"




#ifdef SPATIAL
MyInterval scalarProduct(const MyInterval &x1, const MyInterval &y1, const MyInterval &z1, const MyInterval &x2, const MyInterval &y2, const MyInterval &z2)
{
    return x1 * x2 + y1 * y2 + z1 * z2;
}
#else // PLANAR
MyInterval scalarProduct(const MyInterval &x1, const MyInterval &y1, const MyInterval &x2, const MyInterval &y2)
{
    return x1 * x2 + y1 * y2;
}
#endif

bool isUnique(Count zeroFound)
{
    if ((zeroFound.undecided == 0) && (zeroFound.zeros == 1))
        return true;
    return false;
}

void axialSymOX(const Bodies &bodies, Bodies &mirrorReflection)
/** the reflection with respect the X axis.  x->x, y->-y,z->-z */
{
    MyInterval tempY, tempZ;

    mirrorReflection = bodies;
    for (int i = 0; i < bodies.N - 1; i++)
    {
        tempY = -bodies.y(i);
        mirrorReflection.setY(i, tempY);
#ifdef SPATIAL
        tempZ = -bodies.z(i);
        mirrorReflection.setZ(i, tempZ);
#endif
    }
}

bool colinearCC(const Bodies &bodies, ofstream &outFileTest, ofstream &outFileFailed)
/** if the CC is colinear - whether it is contained in the OX axis */
{
    for (int i = 0; i < bodies.N; i++)
    {
        if (!isSingular(bodies.y(i)))
            return false; // a body has y_i != 0
#ifdef SPATIAL
        if (!isSingular(bodies.z(i)))
            return false; // a body  z_i != 0
#endif
    }

    Bodies mirrorReflection(bodies);
    Bodies hullBodies(bodies);

    axialSymOX(bodies, mirrorReflection);
    intervalHullBodies(bodies, mirrorReflection, hullBodies);

    // the trash variables used for compatybility with the call of searchCC
    //    ofstream outFileTest;
    //    ofstream outFileFailed;
    //    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    //    outFileFailed.open(undecidedTestFile, ios::binary | ios::out | ios::trunc);
    // end of trash

    // 6.10.2020
    bool doSearch = false;
    bool nonDegeneracyTest = false;
    Count zeroFound = searchCC(hullBodies, doSearch, nonDegeneracyTest, outFileTest, outFileFailed); // testing stage

    //    outFileTest.close();
    //    outFileFailed.close();

    if (isUnique(zeroFound))
        return true;
    else
    {
#ifdef CC_SYMMETRY_DEBUG_HELP
        cout << "Couldn't establish uniqueness in collinearity test!" << endl;
#endif
    }
    return blowUpAndProveUniqueSolution(hullBodies, false);
}

#ifdef SPATIAL
MyInterval distFromZeroSq(MyInterval x, MyInterval y, MyInterval z)
/** Returns the distance from zero, squared. */
{
    return (power(x, 2) + power(y, 2) + power(z, 2));
}
#else // PLANAR
MyInterval distFromZeroSq(MyInterval x, MyInterval y)
/** Returns the distance from zero, squared. */
{
    return (power(x, 2) + power(y, 2));
}
#endif

int findGroups(const Bodies &bodies, vector<vector<int>> &vec)
/** On return:
 * `vec[i]`: the vector containing bodies with equal distances from 0, i.e.:
 *     `vec[i][0] = i, vec[i][j] = k` => bodies `i` and `k` have equal distances from 0.
 * Returns the minimal size of the group != 1. If all groups are singletons,
 * returns 0.
 */
{
    int N = bodies.N;
    int minCard = N + 1;
    int card;

    // Create a vector of size N with all values set to 0.
    vector<int> markers(N, 0);

    for (int i = 0; i < N; i++)
    {
        MyInterval xi = bodies.x(i);
        MyInterval yi = bodies.y(i);
#ifdef SPATIAL
        MyInterval zi = bodies.z(i);
        MyInterval distXYZi = distFromZeroSq(xi, yi, zi);
#else // PLANAR
        MyInterval distXYZi = distFromZeroSq(xi, yi);
#endif
        vector<int> row;  // Create an empty row
        row.push_back(i); // Add an element (column) to the row

        card = 1;
        for (int j = 0; j < N; j++)
        {
            if (j == i)
                continue;

            MyInterval xj = bodies.x(j);
            MyInterval yj = bodies.y(j);
#ifdef SPATIAL
            MyInterval zj = bodies.z(j);
            MyInterval distXYZj = distFromZeroSq(xj, yj, zj);
#else // PLANAR
            MyInterval distXYZj = distFromZeroSq(xj, yj);
#endif

            // q(i) and q(j) have the same distance to (0,0)
            if (!emptyIntersection(distXYZi, distXYZj))
            {

                card++;
                row.push_back(j); // Add an element (column) to the row
            }
        }
        vec.push_back(row); // Add the row to the main vector
        if (1 < card && card < minCard)
            minCard = card;
    }

    // all the groups are singletons
    if (minCard == N + 1)
        minCard = 0;
    return minCard;
}

bool symmetricXCC(const Bodies &bodies, ofstream &outFileTest, ofstream &outFileFailed)
//  For equal masses case only
// Returns `true` if the CC is symmetrical with respect to the OX axis (in the planar case) and OXY plane in the spatial case
{

    int NumBodies = bodies.N;

    // we set permutation=identity
    int permutation[NumBodies];
    for (int i = 0; i < NumBodies; i++)
        permutation[i] = i;

    // we try to find a candidate for a possible permutation
    //   which is in agreement with the reflection with repect to the OX axis
    for (int i = 0; i < NumBodies; i++)
    {
        // equalToZero normalization method:
        if (i == NumBodies - 2)
            continue;

#ifdef SPATIAL
        // does not contain 0
        if (!isSingular(bodies.y(i)) || !isSingular(bodies.z(i)))
#else // PLANAR
        if (!isSingular(bodies.y(i))) // doesn't contain 0
#endif
        {
            MyInterval myi = -bodies.y(i);
            MyInterval xi = bodies.x(i);

            for (int j = 0; j < NumBodies; j++)
            {
                if (i == j)
                    continue;
                MyInterval xj = bodies.x(j);
                MyInterval yj = bodies.y(j);
#ifdef SPATIAL
                MyInterval mzi = -bodies.z(i);
                MyInterval zj = bodies.z(j);
                if (!emptyIntersection(xi, xj) && !emptyIntersection(myi, yj) && !emptyIntersection(mzi, zj))
#else // PLANAR
                if (!emptyIntersection(xi, xj) && !emptyIntersection(myi, yj))
#endif
                {
                    permutation[i] = j;
                    permutation[j] = i;
                }
            }
        }
    }
    // now permutation - contains a candidate for symmetry

    // below we construct symBodies --- a set which contains interval hull of bodies and its symmetric image
    Bodies symBodies(bodies);

    MyInterval xi, xj;
    MyInterval yi, yj;
#ifdef SPATIAL
    MyInterval zi, zj;
#endif
    for (int i = 0; i < NumBodies - 2; i++)
    {
        if (permutation[i] == i)
        {
            yi = bodies.y(i);
#ifdef SPATIAL
            zi = bodies.z(i);
#endif

#ifdef PLANAR
            if (isSingular(yi))
            { // if the body lies on the x axis
                symBodies.setY(i, intervalHull(yi, -yi));
            }
#endif
#ifdef SPATIAL
            if (isSingular(yi) && isSingular(zi)) // if the body lies on the x axis
            {
                symBodies.setY(i, intervalHull(yi, -yi));
                symBodies.setZ(i, intervalHull(zi, -zi));
            }
#endif
            else
            {
                cout << "there is no symmetry with respect to the X axis" << endl;
                return false; // because yi doesn't have a "reflection"
            }
        }
        else
        {
            xi = bodies.x(i);
            xj = bodies.x(permutation[i]);
            yi = bodies.y(i);
            yj = bodies.y(permutation[i]);

            symBodies.setX(i, intervalHull(xi, xj));
            symBodies.setY(i, intervalHull(yi, -yj));
#ifdef SPATIAL
            zi = bodies.z(i);
            zj = bodies.z(permutation[i]);

            symBodies.setZ(i, intervalHull(zi, -zj));
#endif
        }
    }

    // the trash variables used for compatybility with the call of searchCC
    //    ofstream outFileTest;
    //    ofstream outFileFailed;
    //    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    //    outFileFailed.open(undecidedTestFile, ios::binary | ios::out | ios::trunc);
    // end of trash

    // 6.10.2020
    bool doSearch = false;
    bool nonDegeneracyTest = false;
    Count zeroFound = searchCC(symBodies, doSearch, nonDegeneracyTest, outFileTest, outFileFailed); // testing stage

    //    outFileTest.close();
    //    outFileFailed.close();

    if (isUnique(zeroFound))
        return true;
    else
    {
#ifdef CC_SYMMETRY_DEBUG_HELP
        cout << "Couldn't establish uniqueness!" << endl;
        cout << "In the test if CC is symmetrical with respect to the OX axis (in the planar case) and OXY plane in the spatial case" << endl;
#endif
    }
    if (blowUpAndProveUniqueSolution(symBodies, false))
        return true;
    else
    {
        cout << "Undecided case even after blowUp!" << endl;
        return false;
    }
    return false;
}

//------------------------------------------------

void transposeMatrix(int size, const MyMatrix &R, MyMatrix &RT)
// on output: RT=R^T
{
    MyMatrix tempRT(size, size);
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            tempRT[i][j] = R[j][i];
        }

    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            RT[i][j] = tempRT[i][j];
        }
}

void writePermCycles(vector<vector<int>> &cycles, int noC)
//  we write to cout the decomposion of permutation into cycles
{
    for (int lc = 0; lc < noC; lc++)
    {
        const int lenCyc = cycles[lc].size();
        if (lc != 0)
            mathCyclesFile << "*";
        mathCyclesFile << "x" << lenCyc;
        cout << "(";
        for (int c = 0; c < lenCyc; c++)
        {
            if (c != 0)
                cout << ",";
            cout << cycles[lc][c];
        }
        cout << ")";
    }
    mathCyclesFile << " + ";
}

void recognizePermCyclesForMathematica(ostream &mathCyclesFile, vector<vector<int>> &cycles,
                                       int noC, int dim)
// we write in Mathematica format the decomposion of permutation into cycles
// noP --- number of permutation for "ksig{noP}"
// dim -- number of bodies
{
    int countCycles[dim + 1];
    // count the number of cycles of particular length
    // coutCycles[i] = number of cycles of length i

    // init coutCycles to 0
    for (int i = 0; i <= dim; i++)
        countCycles[i] = 0;

    int countDiffCycles = 0; // number of different length cycles
    int maxCycleLength = 0;  // maximal length of cycle

    for (int lc = 0; lc < noC; lc++)
    {
        int lenCyc = cycles[lc].size();

        if (maxCycleLength < lenCyc)
            maxCycleLength = lenCyc;

        if (countCycles[lenCyc] == 0) // there was no cycle of length lenCyc before
            countDiffCycles++;

        countCycles[lenCyc]++;
    }

    for (int i = 1; i <= maxCycleLength; i++)
    {
        if (countCycles[i] != 0)
        {
            mathCyclesFile << "x" << i << "^";
            mathCyclesFile << countCycles[i];
        }
    }
    mathCyclesFile << " + ";
}

bool minusIdentity(const MyMatrix &R, int size)
// check if R contains -Id
{
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
        {
            if (i == j && (!R[i][j].contains(-1)))
                return false;

            if (!R[i][j].contains(0))
                return false;
        }
    return true;
}

MyInterval trace(const MyMatrix &R, int dim)
// returns trace(R)
{
    MyInterval tr = 0;
    for (int i = 0; i < dim; i++)
    {
        tr += R[i][i];
    }
    return tr;
}

void permCycles(int permutation[], int N, vector<vector<int>> &cycles, int &noC)
// decomposition of permutation into cycles
// On input:
//    N - the number of elements in permutation
//    permutation[] - the permutation to be decomposed into cycles
// On output:
// cycles[k][0] -- cykl zaczynajacy sie od i
// noC -- liczba cykli --- wspolrzedne w wektorze cycles sa od 0 do (noC-1)
{

    int uI[N]; // uI[i]==0 for elements which are not yet in the cycle decomposition
    for (int i = 0; i < N; i++)
        uI[i] = 0;

    noC = 0;
    for (int i = 0; i < N; i++)
    {
        if (uI[i] != 0) //  the element i is already in some cycle
            continue;

        uI[i] = 1;
        vector<int> row;  // Create an empty row
        row.push_back(i); // Add first element of cycle to the row
        int j, k;
        // the construction of the cycle containing i-th element
        k = i;
        do
        {
            j = permutation[k];
            if (uI[j] == 0) // j-the element was not visited yet, hence j!=i and we add it to the cycle
            {
                row.push_back(j); // Add another element of cycles to the row
                uI[j] = 1;
                k = j;
            }
        } while (j != i);
        cycles.push_back(row); // Add the row to the main vector

        noC++;
    }
}
// */

#ifdef PLANAR
#include "find-sym-2d.cpp"
#endif
#ifdef SPATIAL
#include "find-sym-3d.cpp"
#endif

void writeDataOnSymmetry(int permutation[], int N, const MyMatrix &RT, bool sg)
// writes to cout the data about symmetry
// sg --- sign of det of matrix R:
// sg = TRUE <==> det(R) = 1 <==> R+
// sg = FALSE <==> det(R) = -1 <==> R-
{
    cout << endl;
    cout << "--------------------------------------------------------------------------------" << endl;
    cout << "Permutation:";
    for (int i = 0; i < N; i++)
        cout << " " << permutation[i];
    vector<vector<int>> cycles;
    int noC;
    permCycles(permutation, N, cycles, noC);
    cout << ", cycles: ";
    writePermCycles(cycles, noC);

    //recognizePermCyclesForMathematica(mathCyclesFile, cycles, noC, N);

    //mathCyclesFile << "0" << endl;
    cout << endl;
    cout << "Orthogonal transformation: " << RT << endl;
#ifdef PLANAR
    writesymmetryFromPerm2D(permutation, N, cycles, noC, sg, RT);
#endif
#ifdef SPATIAL
    writesymmetryFromPerm3D(permutation, N, cycles, noC, sg, RT);
#endif
}

bool proveSymmetryWriteToReportFile(const Bodies &bodies, int permutation[], MyMatrix &RT, bool sg,
                                    ofstream &outFileTest, ofstream &outFileFailed)
//  the function proves that (permutation,RT) is a symmetry and writes its description to the report file (cout)
//  bodies - is a box containing a unique CC
//  the symmetry is given as a pair (permutation[],RT)
// // sg --- sign of det of matrix R:
// sg = TRUE <==> det(R) = 1 <==> R+
// sg = FALSE <==> det(R) = -1 <==> R-
//  Return "true" - is the symmetry has been proved
//         "false" - otherwise

{
    int N = bodies.N;
    int invPerm[N];
    for (int i = 0; i < N; i++)
        invPerm[permutation[i]] = i;

    Bodies symBodies(bodies);
    // below we construct symBodies --- a set which contains interval hull of bodies and its symmetric image
    MyInterval RTxi, xj;
    MyInterval RTyi, yj;
#ifdef PLANAR
    MyVector qi(2);
#endif
#ifdef SPATIAL
    MyInterval RTzi, zj;
    MyVector qi(3);
#endif

    for (int i = 0; i < N; i++)
    {
        if (permutation[i] == i)
        {
            continue;
        }
        else
        {
            int j = invPerm[i];
            if (j == N - 1)
                continue;
            xj = bodies.x(j);
            yj = bodies.y(j);
#ifdef SPATIAL
            zj = bodies.z(j);
#endif
            qi[0] = bodies.x(i);
            qi[1] = bodies.y(i);
#ifdef SPATIAL
            qi[2] = bodies.z(i);
#endif

            MyVector RTqi = RT * qi;
            RTxi = RTqi[0];
            RTyi = RTqi[1];
#ifdef SPATIAL
            RTzi = RTqi[2];
#endif

            symBodies.setX(j, intervalHull(RTxi, xj));
            symBodies.setY(j, intervalHull(RTyi, yj));
#ifdef SPATIAL
            symBodies.setZ(j, intervalHull(RTzi, zj));
#endif
        }
    }

    // the trash variables used for compatybility with the call of searchCC
    //    ofstream outFileTest;
    //    ofstream outFileFailed;
    //    outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    //    outFileFailed.open(undecidedTestFile, ios::binary | ios::out | ios::trunc);
    // end of trash

    bool retval = false;
    // 6.10.2020
    bool doSearch = false;
    bool nonDegeneracyTest = false;
    Count zeroFound = searchCC(symBodies, doSearch, nonDegeneracyTest, outFileTest, outFileFailed);

    //    outFileTest.close();
    //    outFileFailed.close();

    // 6.10.2020
    if (zeroFound.undecided == 0)
    {
        if (zeroFound.zeros == 1)
        {
            retval = true;
        }
        else if (zeroFound.zeros > 1)
        {
            cout << "Couldn't establish uniqueness!" << endl;
            return false;
        }
    }

    // the search is inconclusive likely because the set is to small for our accuracy
    // we try to blow up this set
    if (!retval)
    {
        if (blowUpAndProveUniqueSolution(symBodies, false))
        {
            retval = true;
        }
        else
        {
            cout << "undecided case!" << endl;
            return false;
        }
    }

    if (retval)
    {
        writeDataOnSymmetry(permutation, N, RT, sg);
    }
    return retval;
}

bool establishPermutation(const Bodies &bodies, int permutation[], MyMatrix &R)
// for a given ortogonal matrix R the function tries to find a permutation[]
// such that (permutation,R) will be the symmetry of CC contained in bodies
// bodies - is a box containing a unique CC
{

    int N = bodies.N;
    bool found;
    for (int i = 0; i < N; i++)
    {
        found = false;
        /*    if (permutation[i] != -1)
        {
            found = true;
            continue;
        }
*/
#ifdef PLANAR
        MyVector qi(2);
#endif
#ifdef SPATIAL
        MyVector qi(3);
#endif
        qi[0] = bodies.x(i);
        qi[1] = bodies.y(i);
#ifdef SPATIAL
        qi[2] = bodies.z(i);
#endif

#ifdef PLANAR
        MyVector Rqi(2);
#endif
#ifdef SPATIAL
        MyVector Rqi(3);
#endif
        Rqi = R * qi;
        //        cout << "Rq" << i << " = " << Rqi << endl;

        // now we look for pair for qi
        for (int j = 0; j < N; j++)
        {
            MyInterval xj = bodies.x(j);
            MyInterval yj = bodies.y(j);
#ifdef SPATIAL
            MyInterval zj = bodies.z(j);
#endif

            if (!emptyIntersection(Rqi[0], xj) &&
                !emptyIntersection(Rqi[1], yj)
#ifdef SPATIAL
                &&
                !emptyIntersection(Rqi[2], zj)
#endif
                    ) // qj cap Rqi \neq 0
            {
                //              cout << "dla " << i << "znalazl " << j << endl;
                permutation[i] = j;
                found = true;
                break;
            }
        }
        if (!found) // there is no pair for qi body --- this is not proper permutation
        {
            //cout << "there is no pair for q" << i << " body --- this is not proper permutation" << endl;
            return false;
        }
    }

    return true;
}

int FindAndProveAllOrthogonalSymmetries(const Bodies &bodies,
                                        ofstream &outFileTest, ofstream &outFileFailed)
// finds and proves all orthogonal symmetries different from identity
// returns number of  orthogonal symmetries different from identity
{

    int N = bodies.N;

    // the vector containing bodies equidistant from 0
    // vec[i][0] = i, vec[i][j] = k <=> bodies i and k have the same distance from 0
    vector<vector<int>> vec;
    findGroups(bodies, vec);

    int numEqualDist = vec[N - 2].size();

    int sym = 0;
    for (int j = 0; j < numEqualDist; j++) // loops through each body potentially equidistant with q_{n-2}
    {
#ifdef PLANAR
        sym += FindAndProveOrthogonalSymmetriesWithBodyA2D(bodies, vec[N - 2][j],
                                                           outFileTest, outFileFailed);
#endif
#ifdef SPATIAL
        sym += FindAndProveOrthogonalSymmetriesWithBodyA3D(bodies, vec[N - 2][j],
                                                           outFileTest, outFileFailed);
#endif
    }
    return sym;
}




void writeDiffCCandTestSymmetry(int ccNo, vector<Bodies> &ccSolutions, vector<int> &markedSolutions,
                                ofstream &outFileTest, ofstream &outFileFailed)
//  ccNo - the number of entries in ccSolutions, markedSolutions
//      ccSolution[i] is treated as a representant of CC iff markedSolutions[i]==0
//   the function writes to the standard output: different CCs together with the information of their symmetries
//                  for !equalmasses it makes sense to test only the collinearity and flatness
//                   otherwise we check all possible orthogonal transformations
//  the function uses global variables: mathFile,mathListFile
{

    int symColCounter = 0;
    //int symLineCounter = 0;

    int N = ccSolutions[0].N;

    //#ifdef PLANAR
    bool doNotTestColSym = true; // if TRUE, program does not test symmetries for collinear CC
    //#endif

#ifdef SPATIAL
    int symFlatCounter = 0;
    bool doNotTestFlatSym = true; // if TRUE, program does not test symmetries for flat CC
#endif

    cout << endl;
    cout << "Different CC:" << endl;
    int realCC = 0;
    //int counter = 0; // the flag of symmetries by lines:
    // for each body is 1 if there is a symmetry or 0 fi there is no

    for (int i = 0; i < ccNo; i++)
    {
        if (markedSolutions[i] == 0)
        {
            realCC++;
            mathFile << "---------------------" << endl
                     << endl;
            mathFile << "position " << i << endl;
            ccSolutions[i].writeCCformatMath(mathFile, mathListFile);

            cout << "======================================================================================================" << endl;
            cout << "position " << i << endl;
            cout << "======================================================================================================" << endl;
            refineCC(ccSolutions[i]);
            ccSolutions[i].printData(true); // writes distances from 0
            cout << flush;

            writePotAndInertia(ccSolutions[i], false);
            bool symCol = false;
            if (colinearCC(ccSolutions[i], outFileTest, outFileFailed))
            {
                symColCounter++;
                symCol = true;
                cout << "collinear solution no " << symColCounter << endl;
            }

            mathCyclesFile << endl
                           << "(* position " << i << " *)" << endl;
            mathCyclesFile << "ksi" << N << "G" << i << "[";
            for (int j = 1; j < N; j++)
                mathCyclesFile << "x" << j << "_,";
            mathCyclesFile << "x" << N << "_] := (";
            if (symCol)
            {
                mathCyclesFile << "x1^" << N << " + x2^" << N / 2;
                if (N % 2 == 0)
                {
                    mathCyclesFile << ")";
                }
                else
                {
                    mathCyclesFile << "*x1)";
                }
                mathCyclesFile << "/2;" << endl;
                mathCyclesFile << "(* A collinear solution! *)" << endl;
            }

            if (symCol && doNotTestColSym)
                continue;

#ifdef SPATIAL
            bool symFlat = false;
            if (flatCC(ccSolutions[i], outFileTest, outFileFailed))
            {
                symFlatCounter++;
                symFlat = true;
                cout << "planar solution no " << symFlatCounter << endl;
                if (symCol || (symFlat && doNotTestFlatSym))
                    continue;
            }
            else
            {
                MyInterval det;
                if (!detCondForReducedSys(ccSolutions[i], det))
                {
                    cout << "Solution of reduced set of equations, but we cannot claim that this is CC!" << endl;
                    cout << "Determinant is " << det << "singular!" << endl;
                }
            }
#endif
            if (equalMasses)
            {
                int symCount;
                cout << endl;
                symCount = FindAndProveAllOrthogonalSymmetries(ccSolutions[i], outFileTest, outFileFailed);
                mathCyclesFile << " x1^" << N << ")/" << symCount + 1 << ";" << endl; // identity always is a symmetry, thus we add x1^4

                cout << endl
                     << "Number of non-trivial symmetries = " << symCount << endl;
            }
        }
    } // for
    cout << endl;
    cout << "Number of different cc = " << realCC << endl;
}
