#include "defo.h"
#include "test-and-help-functions.h"
#include "estimates-functions.h"
#include "my-interval-func.h"
#include "cute-calc.h"

#include <algorithm>

#include "cc-nb.h"

enum checkZeroResultType
{
    isZeroInside,
    noZeroInside,
    collisionInside
};

bool intersectWithQuadraticBound(MyInterval xSqr, MyInterval xOld, MyInterval &x, MyDouble &ratioNsize)
// this function intend to solve the following problem
//// for some  x \in xOld, we know that x^2 \in xSqr
////// we intend find a new bound for x
/////
// Input:
//// xSqrt = [a,b] contains a  estimate of x^2, we assume that b >= 0!
//// xOld current estimate of x
// Output:
//    x is new estimate of x
//    ratioNsize - the ratio diam(x)/diam(xOld)
// Returns: true  x is a proper subset of xOld
// false otherwise
//
{

    MyInterval bSqrt = sqrt(xSqr.right());
    MyInterval x1;
    if (xSqr.leftBound() <= 0) // if a < 0
    {
        x1 = MyInterval(-bSqrt.rightBound(), bSqrt.rightBound()); // x1=  [-sqrt(b), sqrt(b)]
    }
    else
    {
        MyInterval aSqrt = sqrt(xSqr.left());
        MyInterval temp = MyInterval(aSqrt.leftBound(), bSqrt.rightBound());
        MyInterval minusTemp = -temp; // our quadratic bound is given by minusTemp \cup temp

        MyInterval intersect, minusIntersect;
        bool bIntersect, bMinusIntersect;
        bIntersect = myIntersection(xOld, temp, intersect);                // intersect=xOld \cap temp  // i.e. intersection with positive sqrt(xSqrt)
        bMinusIntersect = myIntersection(xOld, minusTemp, minusIntersect); // minusintersect=xOld \cap minustemp  // i.e. intersection with the negative sqrt(xSqrt)

        if (!bIntersect && !bMinusIntersect)
        {
            cout << "intersectWithQuadraticBound function exit!" << endl;
            exit(1); // impossible! tested in the parent function
        }

        if (bIntersect && bMinusIntersect)
        {
            x1 = intervalHull(intersect, minusIntersect);
        }

        if (!bIntersect && bMinusIntersect)
        {
            x1 = minusIntersect;
        }

        if (bIntersect && !bMinusIntersect)
        {
            x1 = intersect;
        }
    }

    if (!myIntersection(xOld, x1, x))
    {
        cout << "intersectWithQuadraticBound function exit!" << endl;
        exit(1); // impossible! tested in the parent function
    }
    // test if there is improvement
    if ((x.leftBound() > xOld.leftBound()) || (x.rightBound() < xOld.rightBound()))
    {
        ratioNsize = diam(x).rightBound() / diam(xOld).rightBound();
        return true;
    }
    else
        return false;
}

bool distinguishedCubeUEqI(MyInterval U, int i0, int coord, Bodies &bodies)
// exploiting I=U we check if for the  selected coordinate (0-x,1-y,2-z) of the i0-th body  we can obtain tighter bounds - in such case
//   <bodies> are updated
// Returns:
//   true - if we obtain reduction in size by half
// false  - otherwise, even in such case <bodies> might have changed by smaller ratio
{
    MyDouble goodNRatio = 0.5;

    int numBodies = bodies.N;
    MyInterval temp;
    MyDouble ratioNewOld = 0.0; // ratioNewOld\in [0,1)

    if (coord == 1 && i0 == numBodies - 2) // y_{n-2} --- it shouldn't ever happen!
    {
        cout << "Terminated in function distinguishedCubeUEqI" << endl;
        exit(2);
    }
    if (coord == 2 && (i0 == 0 || i0 == numBodies - 2)) // z_0 or z_{n-2}--- it shouldn't ever happen!
    {
        cout << "Terminated in function distinguishedCubeUEqI" << endl;
        exit(2);
    }

    MyInterval reducedI = 0.0;
    for (int j = 0; j < numBodies; j++)
    {
        if (i0 == j)
            continue;
#ifdef PLANAR
        reducedI += bodies.mass(j) * (power(bodies.x(j), 2) + power(bodies.y(j), 2));
#endif
#ifdef SPATIAL
        reducedI += bodies.mass(j) * (power(bodies.x(j), 2) + power(bodies.y(j), 2) + power(bodies.z(j), 2));
#endif
    }
    // reducedI - contains \sum_{j \neq i0} m_jq_j^2  - the moment of inertia without i0-th body

    if (coord == 0) // x- variable
    {
#ifdef SPATIAL
        MyInterval xISqr = (U - reducedI) / bodies.mass(i0) - (power(bodies.y(i0), 2) + power(bodies.z(i0), 2));
#endif
#ifdef PLANAR
        MyInterval xISqr = (U - reducedI) / bodies.mass(i0) - power(bodies.y(i0), 2);
#endif
        if (xISqr < 0)
            return false;

        MyInterval xiOld = bodies.x(i0);
        MyInterval xiOldSqr = power(xiOld, 2);
        if (!myIntersection(xISqr, xiOldSqr, temp))
            return false;

        if (intersectWithQuadraticBound(xISqr, xiOld, temp, ratioNewOld))
        {
            bodies.setX(i0, temp);
            if (ratioNewOld < goodNRatio)
                return true;
        }
    }

    if (coord == 1) // y-variable
    {
#ifdef SPATIAL
        MyInterval yISqr = (U - reducedI) / bodies.mass(i0) - (power(bodies.x(i0), 2) + power(bodies.z(i0), 2));
#endif
#ifdef PLANAR
        MyInterval yISqr = (U - reducedI) / bodies.mass(i0) - power(bodies.x(i0), 2);
#endif
        if (yISqr < goodNRatio)
            return false;

        MyInterval yiOld = bodies.y(i0);
        MyInterval yiOldSqr = power(yiOld, 2);
        if (!myIntersection(yISqr, yiOldSqr, temp))
            return false;

        if (intersectWithQuadraticBound(yISqr, yiOld, temp, ratioNewOld))
        {
            bodies.setY(i0, temp);
            if (ratioNewOld < goodNRatio)
                return true;
        }
    }
#ifdef SPATIAL
    if (coord == 2) // z- variable
    {
        MyInterval zISqr = (U - reducedI) / bodies.mass(i0) - (power(bodies.x(i0), 2) + power(bodies.y(i0), 2));
        if (zISqr < 0)
            return false;

        MyInterval ziOld = bodies.z(i0);
        MyInterval ziOldSqr = power(ziOld, 2);
        if (!myIntersection(zISqr, ziOldSqr, temp))
            return false;

        if (intersectWithQuadraticBound(zISqr, ziOld, temp, ratioNewOld))
        {
            bodies.setZ(i0, temp);
            if (ratioNewOld < goodNRatio)
                return true;
        }
    }
#endif
    return false;
}

bool checkUEqI(Bodies &bodies)
//  Returns false when I=U is incompatible with <bodies>
//   otherwise true is returned and also an attempt is made to reduce <bodies> according with constaint I=U
{

    if (bodies.collisionHappened())
        return true;

    MyInterval U = potential(bodies);
    MyInterval I = momentInertia(bodies);

    if (emptyIntersection(U, I))
        return false; // there is no solution
    else
    {

        int numBodies = bodies.N;

        int dist_i = 0;
        int distCoord = 0;

        bodies.maxDiam(dist_i, distCoord);

        if (!distinguishedCubeUEqI(U, dist_i, distCoord, bodies))
            // the below improvement is costly, so we make only the change is big enough
            // if the change on the smallest interval is less the 50% we stop here
            return true;

        interQuadCounter++; // there was an improvement has we update this counter

        U = potential(bodies);
        for (int i = 0; i < numBodies - 1; i++) // we do not touch the last body, because it is computed from the center of mass
        {
            if (distinguishedCubeUEqI(U, i, 0, bodies)) // x -coordinate
                U = potential(bodies);

            if (i == numBodies - 2)
                continue; // because we drop y_{n-2}, z_{n-2}

            if (distinguishedCubeUEqI(U, i, 1, bodies)) // y -coordinate
                U = potential(bodies);

#ifdef SPATIAL
            if (i == 0)
                continue;                               // we drop z0, hence we do not try for the 0-th body
            if (distinguishedCubeUEqI(U, i, 2, bodies)) // z -coordinate
                U = potential(bodies);
#endif
        }
    }
    return true; // a solution is possible
}
//*******************************

bool vectCollinear(const MyInterval x1, const MyInterval y1, const MyInterval x2, const MyInterval y2)
// returns true if (x1,y1) and (x2,y2) can be collinear
{
    MyInterval vectProd = x1 * y2 - y1 * x2;
    return isSingular(vectProd);
}

#ifdef SPATIAL
bool vectCollinear(const MyInterval x1, const MyInterval y1, const MyInterval z1, const MyInterval x2, const MyInterval y2, const MyInterval z2)
// returns true if (x1,y1,z1) and (x2,y2,z2) can be collinear
{
    if (!vectCollinear(x1, y1, x2, y2)) // z-coordinate of vector product
        return false;

    if (!vectCollinear(x1, z1, x2, z2)) // y-coordinate of vector product
        return false;

    if (!vectCollinear(y1, z1, y2, z2)) // x-coordinate of vector product
        return false;

    return true;
}
#endif

checkZeroResultType checkZeroOneBody(Bodies &bodies, int i, MyDouble &newWidthRel)
// For i-th body we check if there is zero inside  ( fXi,fYi,fZi (bodies))
// returns:
// -- isZeroInside,  // i.e. we cannot rule zero inside and the i-th body do not collide
// -- noZeroInside,
// -- collisionInside // i-the body collides with other
//  On output:
//    if isZeroInside - then we set q_i = Q_i(q_i) \cap q_i - i.e. we use the fixed point formulation of the equation for CC
//    newWidthRel -- relative shrinkage of change of volume of q_i = newVolume/oldVolume
{

    newWidthRel = 1.0;
    MyInterval fXi, fYi, gXi, gYi;
#ifdef SPATIAL
    MyInterval fZi, gZi;
#endif
    MyInterval old;

#ifdef PLANAR
    if (bodies.fXY(i, fXi, gXi, fYi, gYi)) // q_i does not interfere (do not have possible collision) with any other body
#endif
#ifdef SPATIAL
        if (bodies.fXYZ(i, fXi, gXi, fYi, gYi, fZi, gZi)) // q_i does not interfere  (do not have possible collision)  with any other body
#endif
        {

            if (!isSingular(fXi) || !isSingular(fYi)
#ifdef SPATIAL
                || !isSingular(fZi)
#endif
                    ) // this is checking whether 0 \in f_{qi}
            {
                return noZeroInside;
            }
            else
            {
#ifdef PLANAR
                if (!vectCollinear(gXi, gYi, bodies.x(i), bodies.y(i))) // there is no effect for 4 bodies // point-force collinearity test fails
                    return noZeroInside;
#endif
#ifdef SPATIAL
                if (!vectCollinear(gXi, gYi, gZi, bodies.x(i), bodies.y(i), bodies.z(i))) //  point-force collinearity test fails
                    return noZeroInside;
#endif
                // an attempt to improve using the fixed point formulation of the equation q_i=g(q)

                if (i != bodies.N - 1) // the last body is computed from the center of mass, hence we do not work with it
                {
                    MyInterval x = bodies.x(i);
                    old = x;
                    if (!myIntersection(x, gXi, x))
                        return noZeroInside;
                    newWidthRel = diam(x).rightBound() / diam(old).rightBound();
                    bodies.setX(i, x);

                    if (i != bodies.N - 2) // y_{n-2} is dropped
                    {
                        MyInterval y = bodies.y(i);
                        old = y;
                        if (!myIntersection(y, gYi, y))
                            return noZeroInside;
                        newWidthRel *= diam(y).rightBound() / diam(old).rightBound();
                        bodies.setY(i, y);
                    }

#ifdef SPATIAL
                    if ((i != bodies.N - 2) && (i != 0)) // z_0,z_{n-2} is dropped
                    {
                        MyInterval z = bodies.z(i);
                        old = z;
                        if (!myIntersection(z, gZi, z))
                            return noZeroInside;
                        newWidthRel *= diam(z).rightBound() / diam(old).rightBound();
                        bodies.setZ(i, z);
                    }
#endif
                }

                return isZeroInside;
            }
        }
        else // there is collision
        {
            return collisionInside;
        }
}

bool checkZero(Bodies &bodies)
// return false, if some i-th  we establish that 0 \notin f_{q_i} for all q \in bodies
//    otherwise true is returned
//  bodies - are refined by some iteration procedure
{
    int numBodies = bodies.N;

    MyDouble widthRel;

    //first check for the longestInterval
    //if there is no satisfactory improvement here, skip the test
    int min_i = 0;
    int minCoord = 0;
    // bodies.minDiam(min_i, minCoord); // this is slighly worse
    bodies.maxDiam(min_i, minCoord);

    widthRel = 1.0;
    if (checkZeroOneBody(bodies, min_i, widthRel) == noZeroInside)
        return false; // there is no solution here

    if (widthRel < 1.0) // this means that there was an improvement for the selected coordinate
    {
        int loop = 1;
        do
        {
            widthRel = 1.0;
            for (int i = 0; i < numBodies - 1; i++)
            {
                if (loop == 1 && i == min_i)
                {
                    loop++;
                    continue;
                }

                MyDouble newWidthRel = 1.0;
                if (checkZeroOneBody(bodies, i, newWidthRel) == noZeroInside)
                    return false; // there is no solution here

                widthRel *= newWidthRel;
            }
        } while (widthRel < 0.1);
    }
    else // there was no improvement for the selected coordinate, we go through all bodies (exept  min_i and n-1)
    {
        for (int i = 0; i < numBodies - 1; i++)
        {
            MyDouble newWidthRel = 1.0;
            if (i == min_i)
                continue;

            if (checkZeroOneBody(bodies, i, newWidthRel) == noZeroInside)
                return false; // there is no solution here
        }
    }
    return true;
}

bool distanceTest(const Bodies &bodies)
// returns TRUE if:
//    when some constraints on mutual possition of the bodies are satisfied
//
// q_{n-2} is the body furthest from the origin
// and x_{n-2} > 0, y_{n-2} = 0,z_{n-2}

// IN THE PLANAR CASE
// y0 >= 0
//  additionaly in the case of equalMasses or !fixedOrderOfBodies
//    x0 <= xi for all xi  i.e. q0 the most left body
//    y1 <= yi for all yi (if N > 3) i.e. y1 the smallest y in the bodies
// IN THE SPATIAL CASE
//  // q_0 is the body furthest from the OX line
// y0 >= 0
// q_1 is the body furthest from the OXY plane
// z1 >= 0

/////  FOR THE REMAINING BODIES in the case of equalMasses
//  for the remaining bodies:
//    if (orderIncreasing)  x2 <= x3 <= ... <= x_{n-3} <= x_{n-1}
//     else   x2 >= ... >= x_{n-3} >= x_{n-1}
//
{
    int numBodies = bodies.N;

#ifdef PLANAR
    MyInterval dist2;
    if (bodies.y(0) < 0)
        return false; // y0 has to be nonnegative!
    if (bodies.x(numBodies - 2) <= 0)
        return false; // q_{N-2} has to be furthest to the right!

    MyInterval xFurthestSqr = power(bodies.x(numBodies - 2), 2); // y_{N-2} = 0
    // we verify whether q_{n-2} is the body with the largest distance from the origin
    if (equalMasses || !fixedOrderOfBodies)
    {
        for (int i = 0; i < numBodies; i++)
        {
            if (i != numBodies - 2)
            {
                dist2 = power(bodies.x(i), 2) + power(bodies.y(i), 2);
                if (dist2.leftBound() > xFurthestSqr.rightBound())
                    return false; // (N-2)-th body has to be furthest
            }
        }
    }
#endif
#ifdef SPATIAL
    if (bodies.y(0) < 0)
        return false; // y0 has to be nonnegative!
    if (bodies.x(numBodies - 2) <= 0)
        return false; // q_{N-2} from the origin and on the OX axis with positive coordinate
    if (bodies.z(1) < 0)
        return false; // z_1 should be maximal for all |z_i|

    MyInterval distx2, disty2, distz2;
    //checking whether q_{n-2} is the furthest body from the origin
    //                 q_0 - is the furthest body from OX
    //                 q_1 - is the furtherst body from OXY
    MyInterval xFurthestSqr = power(bodies.x(numBodies - 2), 2);
    MyInterval yzFurthestSqr = power(bodies.y(0), 2);
    MyInterval z1Square = power(bodies.z(1), 2);
    for (int i = 0; i < numBodies; i++)
    {
        if (i != numBodies - 2)
        {
            distx2 = power(bodies.x(i), 2);
            disty2 = power(bodies.y(i), 2);
            distz2 = power(bodies.z(i), 2);

            if ((distx2 + disty2 + distz2).leftBound() > xFurthestSqr.rightBound())
                return false; // (N-2)-th body has to be furthest

            if ((disty2 + distz2).leftBound() > yzFurthestSqr.rightBound())
                return false; // q0 is has to be the furthest from OX axis

            if (distz2.leftBound() > z1Square.rightBound())
                return false; // q1 is has to be the furthest from OXY plane
        }
    }
#endif

    // for equal masses we demand the particular order
    if (equalMasses)
    {
#ifdef PLANAR
        // x0 <= xi for all xi
        MyInterval xq0 = bodies.x(0).leftBound();
        for (int i = 1; i < numBodies; i++)
        {
            if (bodies.x(i) < xq0)
                return false;
        }
        // y1 <= yi for all yi (if N > 3)
        if (numBodies > 3) // when numBodies =3  this does not make sence because q_1=q_{n-2}
        {
            MyInterval yq1 = bodies.y(1).leftBound();
            for (int i = 0; i < numBodies; i++)
            {
                if (i == 1)
                    continue;

                if (bodies.y(i) < yq1)
                    return false;
            }
        }
#endif

        if (numBodies > 4) // for numBodies <=4 this does not make sense since q_0,q_1,q_2 - are "fixed" already by other demands
        {
            if (orderIncreasing)
            {
                // x2 <= x3 <= ... <= x_{n-3} <= x_{n-1}
                MyInterval x = bodies.x(2).leftBound();
                MyInterval xNext;
                for (int i = 3; i < numBodies; i++)
                {
                    if (i == numBodies - 2)
                        continue;

                    xNext = bodies.x(i);
                    if (xNext < x)
                        return false;

                    x = xNext.leftBound();
                }
            }
            else
            {
                // x2 >= ... >= x_{n-3} >= x_{n-1}
                MyInterval x = bodies.x(numBodies - 1).leftBound();
                MyInterval xNext;
                for (int i = numBodies - 3; i > 1; i--)
                {
                    xNext = bodies.x(i);
                    if (xNext < x)
                        return false;

                    x = xNext.leftBound();
                }
            }
        }
    }
    return true; // the solution is possible
}

bool intersectAndImprovement(MyInterval xOld, MyInterval &x, bool &intersectionEmpty)
// On output:
//     x=x \cap xOld
//    true is returned when on output x \subset int xOld
{
    intersectionEmpty = false;
    if (!myIntersection(xOld, x, x))
    {
        intersectionEmpty = true;
        return false;
    }
    //  x(out)=x(in) \cap xOld
    return (x.leftBound() > xOld.leftBound()) || (x.rightBound() < xOld.rightBound());
}

void piszLocalData(int i0, vector<int> &cluster, MyInterval xiOld, MyInterval xi0, const Bodies &bodies)
//  the ad-hoc function to report errors with <refineBasedZeroInCluster>
{
    int sizeCluster = cluster.size();

    cout << "cluster size=" << sizeCluster << endl;
    cout << "cluster: ";
    for (int j0 = 0; j0 < sizeCluster; j0++)
        cout << "  " << cluster[j0];
    cout << endl;

    cout << "problematic body=" << i0 << endl;
    cout << "varold=" << xiOld << "   xnew=" << xi0 << endl;

    bodies.printData();
    cout << endl;

    bodies.printDiam();
    cout << endl;
}

#ifdef PLANAR
void computeSumMiQiOnCluster(const Bodies &bodies, vector<int> &cluster, int iNotIncluded, MyInterval &smqX, MyInterval &smqY)
#endif
#ifdef SPATIAL
    void computeSumMiQiOnCluster(const Bodies &bodies, vector<int> &cluster, int iNotIncluded, MyInterval &smqX, MyInterval &smqY, MyInterval &smqZ)
#endif
// computation of   \sum_{j \in cluster, j\neq iNotIncluded} m_j q_j ={smqX,smqY,smqZ}
{

    smqX = 0;
    smqY = 0;
#ifdef SPATIAL
    smqZ = 0;
#endif
    MyInterval mi;
    int sizeCluster = cluster.size();
    for (int j = 0; j < sizeCluster; j++)
    {
        int i = cluster[j];
        if (i == iNotIncluded)
            continue;
        mi = bodies.mass(i);
        smqX += mi * bodies.x(i);
        smqY += mi * bodies.y(i);
#ifdef SPATIAL
        smqZ += mi * bodies.z(i);
#endif
    }
}

#ifdef PLANAR
void computeExternalForceOnCluster(const Bodies &bodies, vector<int> &cluster, MyInterval &FExtX, MyInterval &FExtY)
#endif
#ifdef SPATIAL
    void computeExternalForceOnCluster(const Bodies &bodies, vector<int> &cluster, MyInterval &FExtX, MyInterval &FExtY, MyInterval &FExtZ)
#endif
// computation \sum_{j \in cluster,k notin cluster } m_j * m_k * (q_j-q_k)/r^{3}_{jk}={FExtX,FExtY,FExtZ} // the sum of MINUS external force acting on the cluster
{
    int N = bodies.N;
    MyInterval qx, qy, qz;
    FExtX = 0;
    FExtY = 0;
#ifdef SPATIAL
    FExtZ = 0;
#endif

    int sizeCluster = cluster.size();
    for (int j = 0; j < sizeCluster; j++)
    {
        int i = cluster[j];
        MyInterval mi = bodies.mass(i);
        MyInterval tempX = 0.0;
        MyInterval tempY = 0.0;
        MyInterval tempZ = 0.0;
        for (int k = 0; k < N; k++)
        {
            if (isInCluster(cluster, k))
                continue;
// now we know that k is not in cluster
#ifdef PLANAR
            cuteCalculation(bodies.x(i) - bodies.x(k), bodies.y(i) - bodies.y(k), qx, qy, 3, 1); // qx=(x_i - x_k)/r_{ik}^3,  qy=(y_i - y_k)/r_{ik}^3
#endif
#ifdef SPATIAL
            cuteCalculation(bodies.x(i) - bodies.x(k), bodies.y(i) - bodies.y(k), bodies.z(i) - bodies.z(k),
                            qx, qy, qz, 3, 1); // qx=(x_i - x_k)/r_{ik}^3,  qy=(y_i - y_k)/r_{ik}^3, qz=(z_i - z_k)/r_{ik}^3
#endif

            MyInterval mk = bodies.mass(k);
            MyInterval mimk = mi * mk;
            tempX += mimk * qx;
            tempY += mimk * qy;
#ifdef SPATIAL
            tempZ += mimk * qz;
#endif
        } // end for k

        FExtX += tempX;
        FExtY += tempY;
#ifdef SPATIAL
        FExtZ += tempZ;
#endif
    } // end for j
}

#ifdef PLANAR
bool refineBasedZeroInCluster(Bodies &bodies, vector<int> &cluster, MyInterval FExtCx, MyInterval FExtCy)
#endif
#ifdef SPATIAL
    bool refineBasedZeroInCluster(Bodies &bodies, vector<int> &cluster, MyInterval FExtCx, MyInterval FExtCy, MyInterval FExtCz)
#endif
// function tries to refine <Bodies> based on the zero-test for the <cluster>
//  On input: (FExtCx,  FExtCy, FExtCz)=\sum_{j \in cluster,k notin cluster } m_j*m_k* (q_j-q_k)/r^{3}_{jk}, the sum of external forces on the cluster
// Returns False - if the possiblity of zero has been excluded
//                 otherwise returns true
{

    int N = bodies.N;
    int sizeCluster = cluster.size();
    for (int j0 = 0; j0 < sizeCluster; j0++)
    {
        int i0 = cluster[j0];

        if (i0 == N - 1)
            continue; // q_{N-1} is an dependent variable, q_{N-1} is obtained from the center of mass condition

        MyInterval smx = 0.0;
        MyInterval smy = 0.0;
        MyInterval smz = 0.0;
        // computation of \sum_{j \in cluster, j \neq j0} m_j q_j ={smx,smy,smz}
#ifdef PLANAR
        computeSumMiQiOnCluster(bodies, cluster, i0, smx, smy);
#endif
#ifdef SPATIAL
        computeSumMiQiOnCluster(bodies, cluster, i0, smx, smy, smz);
#endif

        bool intersectionEmpty = false;

        MyInterval xi0Old = bodies.x(i0);
        MyInterval xi0 = (FExtCx - smx) / bodies.mass(i0); // the constraint coming from considering the cluster
        if (intersectAndImprovement(xi0Old, xi0, intersectionEmpty))
        {
            bodies.setX(i0, xi0);
            clusterImprovementCounter++;
        }
        if (intersectionEmpty)
        {
            return false;
        }

        if (i0 != N - 2) // because y_{n-2}=0
        {
            MyInterval yi0Old = bodies.y(i0);
            MyInterval yi0 = (FExtCy - smy) / bodies.mass(i0);
            if (intersectAndImprovement(yi0Old, yi0, intersectionEmpty))
            {
                bodies.setY(i0, yi0);
                clusterImprovementCounter++;
            }

            if (intersectionEmpty)
            {
                return false;
            }
        }

#ifdef SPATIAL
        if ((i0 != N - 2) && (i0 != 0))
        { // because z_{n-2}=0, z_0=0
            MyInterval zi0Old = bodies.z(i0);
            MyInterval zi0 = (FExtCz - smz) / bodies.mass(i0);
            if (intersectAndImprovement(zi0Old, zi0, intersectionEmpty))
            {
                bodies.setZ(i0, zi0);
                clusterImprovementCounter++;
            }
            if (intersectionEmpty)
            {
                return false;
            }
        }
#endif
    }
    return true;
}

bool zerosInGroup(Bodies &bodies, vector<int> &cluster /*,  bool testMode = false */)
// this is the zero test for the <cluster>
// returns FALSE when the solution is not possible based on the zero-test
// otherwise returns TRUE
//  in the normal mode of action testMode=false
//  if testMode=true then we just write a lot of info to the file
{

    MyInterval FExtX = 0.0;
    MyInterval FExtY = 0.0;
    MyInterval FExtZ = 0.0;

#ifdef PLANAR
    computeExternalForceOnCluster(bodies, cluster, FExtX, FExtY);
#endif
#ifdef SPATIAL
    computeExternalForceOnCluster(bodies, cluster, FExtX, FExtY, FExtZ);
#endif
    // now: \sum_{j \in cluster,k notin cluster } m_j m_k (q_j-q_k)/r^{3}_{jk}={FExtX,FExtY,FExtZ}

    // computation of \sum_{j \in cluster} m_j q_j ={smx,smy,smz}
    MyInterval smx = 0.0;
    MyInterval smy = 0.0;
    MyInterval smz = 0.0;

    int outOfCluster = -1; // any values outside of <cluster> is ok
#ifdef PLANAR
    computeSumMiQiOnCluster(bodies, cluster, outOfCluster, smx, smy);
#endif
#ifdef SPATIAL
    computeSumMiQiOnCluster(bodies, cluster, outOfCluster, smx, smy, smz);
#endif

#ifdef TESTY
    /*
    if(testMode){ // we write some info for debugging purpose
	  cout << "smx=" << smx << "  smy=" << smy;
	  #ifdef SPATIAL
	  cout << "  smz=" << smz;
	  #endif
	  cout << endl << "FExtX="<< FExtX << "  FExtY=" << FExtY ;
	  #ifdef SPATIAL
	   cout  << "  FExtZ="<< FExtZ;
	  #endif
	  cout << endl;
	  cout << "smx - FExtX=" << smx - FExtX << "   smy - FExtY=" << smy - FExtY;
	  #ifdef SPATIAL
	  cout << "   smz - FExtZ=" << smz - FExtZ  ;
	  #endif
	  cout << endl;

	}
	*/
#endif

    if (!isSingular(smx - FExtX) || !isSingular(smy - FExtY)
#ifdef SPATIAL
        || !isSingular(smz - FExtZ)
#endif
    )
    {
        return false; // there is NO solution
    }

    bool retval = true;
// now an attempt to refine <bodies>
//if(!testMode){
#ifdef SPATIAL
    retval = refineBasedZeroInCluster(bodies, cluster, FExtX, FExtY, FExtZ);
#endif
#ifdef PLANAR
    retval = refineBasedZeroInCluster(bodies, cluster, FExtX, FExtY);
#endif
    //}

    return retval;
}

bool clusterTest(Bodies &bodies)
// we perform the zero test and the test based on the balance of moment of inertia and potential for the clusters of bodies
// the zero test sometimes shrinks <bodies>
// on the output:
// false -- there is NO solution
// true -- solution is POSSIBLE
{

    vector<int> cluster;
    MyInterval ICZ, UCZ, FCZ;
    int N = bodies.N;

    MyDouble eps1 = 1.0 / (MyDouble(N) * N * (N - 1) * (N - 1)); // (n-1)^2 n^2 - estimate from below for the smallest distance for the equal mass case
    MyDouble delta = eps1 * 50;

    for (int i = 0; i < N; i++) // cluster contains all bodies
        cluster.push_back(i);

    UCZ = infPotEnergyCluster(bodies, cluster);
    ICZ = supMomInertiaCluster(bodies, cluster);

    if (UCZ > ICZ)
    {
        return false; // there is NO solution
    }

    // the vector bodiesInCluster allows to run functions in clusterTest only once for each body;
    // i.e. if cluster = <0, 2, 3>, then we run functions only once for i = 0
    vector<int> bodiesInCluster; // bodiesInCluster[i]==0 - means that i-th body was not tested
                                 // if  bodiesInCluster[i]==1 -means that i-th body has been already tested
    for (int i = 0; i < N; i++)  // initialization of bodiesInCluster
        bodiesInCluster.push_back(0);

    for (int i = 0; i < N - 1; i++)
    { // for each body we find a cluster containing it
        if (bodiesInCluster[i] == 1)
            continue; // body qi belongs to another cluster, which was already tested

        if (findCluster(bodies, i, delta, cluster)) // true if cluster contains at least TWO bodies
        {
            if ((int)cluster.size() < N) // cluster does not contain all bodies,  projector (int) is used to avoid compiler warning
            {
                UCZ = infPotEnergyCluster(bodies, cluster);
                ICZ = supMomInertiaCluster(bodies, cluster);
                FCZ = infFCZ(bodies, cluster);

                if (UCZ + FCZ > ICZ)
                    return false; // there is NO solution, based on the general balance of mom of inertia, pot energy and the force from
                                  // bodies not in the cluster

                if (!zerosInGroup(bodies, cluster))
                    return false; // this is the zero for the cluster
            }
        }
        updateVectorBodiesInCluster(bodiesInCluster, cluster); // we mark all bodies in cluster
    }

    return true; // solution is POSSIBLE
}

MyInterval aprioriEstimate(int N)
// returns the upper bound for CC with N-bodies
{
    if (N <= 4)
        return MyInterval(N - 1);
    else
    {
        MyInterval two = 2.0;
        MyInterval e1 = MyInterval(1.0) / 3;
        MyInterval e2 = MyInterval(-2.0) / 3;
        MyInterval e3 = MyInterval(2.0) / 3;
        return (power(two, e1) +
                power(two, e2)) *
               power(MyInterval(N - 2), e3);
    }
}

bool checkAprioriBounds(const Bodies &bodies)
// returns FALSE --- apriori bounds are violated
//  true - otherwise
{
    int N = bodies.N;
    MyInterval aE2 = power(aprioriEstimate(N), 2);

    for (int i = 0; i < N; i++)
    {
#ifdef PLANAR
        if (power(bodies.x(i), 2) + power(bodies.y(i), 2) > aE2)
#endif
#ifdef SPATIAL
            if (power(bodies.x(i), 2) + power(bodies.y(i), 2) + power(bodies.z(i), 2) > aE2)
#endif
                return false;
    }
    return true;
}

bool possibleSolution(Bodies &bodies)
//  Input values:
//     bodies - here we try to exclude CC
// returns: true if are tests rule out the existence of CC in <bodies>
//          false otherwise
//         some of the test update (shrink) bodies, so that all possible CC in bodies(on output) are contained
//         in bodies (the output value)
{
    // test 1 ----------------------------------------------------------------------
    if (!checkAprioriBounds(bodies))
    {
        aprioriCounter++;
        return false; // there is no solution here
    }

    // test 2 ----------------------------------------------------------------------
    if (!distanceTest(bodies))
    {
        distanceCounter++;
        return false; // there is no solution here
    }

    // test 3 ----------------------------------------------------------------------
    // naive interval evaluation of f_i(bodies); for cc it must be that 0 in f_i(bodies)
    if (!checkZero(bodies))
    {
        checkZeroCounter++;
        return false; // there is no solution here
    }

    // test 4----------------------------------------------------------------------
    if (!checkUEqI(bodies))
    {

        UEqICounter++;
        return false;
    }

    // test 5 ----------------------------------------------------------------------
    if (!clusterTest(bodies))
    {
        clusterCounter++;
        return false; // there is no solution here
    }

    return true;
}

bool detCondForReducedSys(const Bodies &bodies, MyInterval &det)
/** returns TRUE if determinant of the matix A:
 *  (y_0 - y_{n-1})   (y_{n-2} - y_{n-1})
 * -(x_0 - x_{n-1})  -(x_{n-2} - x_{n-1})
 * is not singular, i.e. 0\not\in det A
 * bodies is CC
*/
// this in 3D is needed to claim that the solution of the reduced system is in fact a central configuration
{
    int N = bodies.N;
    MyInterval x0 = bodies.x(0);
    MyInterval y0 = bodies.y(0);
    MyInterval xBL = bodies.x(N - 2);
    MyInterval yBL = bodies.y(N - 2);
    MyInterval xL = bodies.x(N - 1);
    MyInterval yL = bodies.y(N - 1);

    det = (x0 - xL) * (yBL - yL) - (y0 - yL) * (xBL - xL);
    if (!isSingular(det))
        return true;
    return false;
}

//-----------------------
void setAprioriBoundsData(Bodies &bodies)
//  set general appriori bounds, both for equal and non-equal masses case
//  the function should be in agreement with <distanceTest>
{
    int N = bodies.N;
    MyInterval intervalOverSize = MyInterval(-1e-4, 9.0 * 1e-4);
    MyInterval a = aprioriEstimate(N);
    MyInterval aBound = intervalHull(-a, a);
    for (int i = 0; i < N - 1; i++)
    {
        bodies.setX(i, aBound + intervalOverSize);
        bodies.setY(i, aBound + intervalOverSize);
#ifdef SPATIAL
        bodies.setZ(i, aBound + intervalOverSize);
#endif

        if (i == N - 2) // we drop y_{n-2} and z_{n-2}, and  the body is furtherst to right
        {
            bodies.setX(i, intervalHull(MyInterval(1) / 2, a) + intervalOverSize);
            bodies.setY(i, 0.0);
#ifdef SPATIAL
            bodies.setZ(i, 0.0);
#endif
        }
#ifdef PLANAR
        if (i == 0) // q0 is furthers to the left and y0\geq 0
        {
            bodies.setX(i, intervalHull(MyInterval(0), -a) + intervalOverSize); // x0 must be negative
            bodies.setY(i, intervalHull(MyInterval(0), a) + intervalOverSize);  // y0 \geq 0
        }
#endif
#ifdef SPATIAL
        if (i == 0) // q0 is furtherst from the line OX with z0=0, y0\geq 0
        {
            bodies.setY(i, intervalHull(MyInterval(0), a) + intervalOverSize);
            bodies.setZ(i, 0.0);
        }
        if (i == 1) // q1 is  furtherst from the plane OXY with z1 \geq 0
        {
            bodies.setZ(i, intervalHull(MyInterval(0), a) + intervalOverSize); // z1 \geq 0
        }
#endif
    }
}

MyInterval dfdxNormL1(MyMatrix &DFF, int j, int size)
//funkcja zwraca normę L1 kolumny j-tej w macierzy DFF
{
    MyInterval dfdxNorm = MyInterval(0.0);
    for (int i = 0; i < size; i++)
    {
        dfdxNorm += abs(DFF[i][j]);
    }
    return dfdxNorm;
}

MyInterval dfdxNormMax(MyMatrix &DFF, int j, int size)
//funkcja zwraca normę Max kolumny j-tej w macierzy DFF
{
    MyInterval temp = MyInterval(0.0);
    MyInterval dfdxNorm = MyInterval(0.0);
    for (int i = 0; i < size; i++)
    {
        temp = abs(DFF[i][j]);
        if (temp > dfdxNorm)
        {
            dfdxNorm = temp;
        }
    }
    return dfdxNorm;
}

// 6.10.2020
MyDouble midPointInt(MyInterval interval)
{
    return (interval.leftBound() + interval.rightBound()) / 2;
}

#ifdef PLANAR
void fillAuxMatrix(const Bodies &bodies, MyDouble dpmiN[][3])
#endif
#ifdef SPATIAL
    void fillAuxMatrix(const Bodies &bodies, MyDouble dpmiN[][6])
#endif
{
    // auxiliary matrix to remember constant parts for i:
    // dpmiN[i][0] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(x_i - x_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][1] = 3 \frac{(x_i - x_(N-1))(y_i - y_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][2] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(y_i - y_(N-1))^2}{r_{i(N-1)}^5}
    // #ifdef SPATIAL
    // dpmiN[i][3] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(z_i - z_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][4] = 2 \frac{(x_i - x_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][5] = 3 \frac{(y_i - y_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}

    int N = bodies.N;
    MyInterval r;
    MyDouble r3, r5; //  r3=r^{-3}, r5=r^{-5}
    MyDouble dx, dy;
#ifdef SPATIAL
    MyDouble dz;
#endif

    for (int i = 0; i < N - 1; i++)
    {
        r = bodies.radius(i, N - 1);
        r3 = midPointInt(power(r, -3));
        r5 = midPointInt(power(r, -5));
        dx = midPointInt(bodies.x(i) - bodies.xN());
        dy = midPointInt(bodies.y(i) - bodies.yN());
#ifdef SPATIAL
        dz = midPointInt(bodies.z(i) - bodies.zN());
#endif

        dpmiN[i][0] = r3 - 3 * dx * dx * r5;
        dpmiN[i][1] = 3 * dx * dy * r5;
        dpmiN[i][2] = r3 - 3 * dy * dy * r5;
#ifdef SPATIAL
        dpmiN[i][3] = r3 - 3 * dz * dz * r5;
        dpmiN[i][4] = 3 * dx * dz * r5;
        dpmiN[i][5] = 3 * dy * dz * r5;
#endif

    } // for
}

#ifdef PLANAR
MyDouble dfdmjPLANAR(const Bodies &bodies, int j, MyDouble dpmiN[][3])
/*
returns |df/dm_j|*diam(m_j) =  \sum_{i = 0}^{N-2}\sum_{k=1}^{2} |df_i^2/dm_j|* diam(m_j),

    // auxiliary matrix to remember constant parts for i:
    // dpmiN[i][0] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(x_i - x_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][1] = 3 \frac{(x_i - x_(N-1))(y_i - y_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][2] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(y_i - y_(N-1))^2}{r_{i(N-1)}^5}

*/
{
    int N = bodies.N;
    MyDouble tempSum = 0.0;
    for (int i = 0; i < N - 1; i++)
    {
        MyInterval r = bodies.radius(i, N - 1);
        MyDouble r3 = midPointInt(power(r, -3));
        MyDouble d = midPointInt(diam(bodies.x(j)));

        // tempSum = |f_i^x|*diam(m_j)
        tempSum += midPointInt(bodies.x(i) - bodies.x(j)) * r3; // (x_i - x_j)/r_{ij}^3
        tempSum -= midPointInt(bodies.x(j) + bodies.xN()) * dpmiN[i][0];
        tempSum += midPointInt(bodies.y(j) + bodies.yN()) * dpmiN[i][1];
        tempSum = abs(tempSum) * d;

        // tempSum = tempSum + |f_i^y|*diam(m_j)
        tempSum += midPointInt(bodies.y(i) - bodies.y(j)) * r3; // (y_i - y_j)/r_{ij}^3
        tempSum += midPointInt(bodies.x(j) + bodies.xN()) * dpmiN[i][1];
        tempSum -= midPointInt(bodies.y(j) + bodies.yN()) * dpmiN[i][2];
        tempSum = abs(tempSum) * d;
    }

    return tempSum;
} //dfdmjPLANAR
#endif

#ifdef SPATIAL
MyDouble dfdmjSPATIAL(const Bodies &bodies, int j, MyDouble dpmiN[][6])
/*
returns |df/dm_j|*diam(m_j) =  \sum_{i = 0}^{N-2}\sum_{k=1}^{3} |df_i^3/dm_j|* diam(m_j),

    // auxiliary matrix to remember constant parts for i:
    // dpmiN[i][0] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(x_i - x_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][1] = 3 \frac{(x_i - x_(N-1))(y_i - y_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][2] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(y_i - y_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][3] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(z_i - z_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][4] = 2 \frac{(x_i - x_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][5] = 3 \frac{(y_i - y_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}

*/
{
    int N = bodies.N;
    MyDouble tempSum = 0.0;
    for (int i = 0; i < N - 1; i++)
    {
        MyInterval r = bodies.radius(i, N - 1);
        MyDouble r3 = midPointInt(power(r, -3));
        MyDouble d = midPointInt(diam(bodies.mass(j)));

        // tempSum = |f_i^x|*diam(m_j)
        tempSum += midPointInt(bodies.x(i) - bodies.x(j)) * r3; // (x_i - x_j)/r_{ij}^3
        tempSum -= midPointInt(bodies.x(j) + bodies.xN()) * dpmiN[i][0];
        tempSum += midPointInt(bodies.y(j) + bodies.yN()) * dpmiN[i][1];
        tempSum += midPointInt(bodies.z(j) + bodies.zN()) * dpmiN[i][4];
        tempSum = abs(tempSum) * d;

        // tempSum = tempSum + |f_i^y|*diam(m_j)
        tempSum += midPointInt(bodies.y(i) - bodies.y(j)) * r3; // (y_i - y_j)/r_{ij}^3
        tempSum += midPointInt(bodies.x(j) + bodies.xN()) * dpmiN[i][1];
        tempSum -= midPointInt(bodies.y(j) + bodies.yN()) * dpmiN[i][2];
        tempSum += midPointInt(bodies.z(j) + bodies.zN()) * dpmiN[i][5];
        tempSum = abs(tempSum) * d;

        // tempSum = tempSum + |f_i^z|*diam(m_j)
        tempSum += midPointInt(bodies.z(i) - bodies.z(j)) * r3; // (y_i - y_j)/r_{ij}^3
        tempSum += midPointInt(bodies.x(j) + bodies.xN()) * dpmiN[i][4];
        tempSum += midPointInt(bodies.y(j) + bodies.yN()) * dpmiN[i][5];
        tempSum -= midPointInt(bodies.z(j) + bodies.zN()) * dpmiN[i][3];
        tempSum = abs(tempSum) * d;
    }

    return tempSum;
}
#endif

MyDouble dxdMjNorm(const Bodies &bodies, int &jVar)
/*
returns max_{j\in {0, ..., N-2}} |df/dm_j|*diam(m_j) = max_{j\in {0, ..., N-2}} \sum_{i = 0}^{N-2}\sum_{k=1}^{d} |df_i^d/dm_j|* diam(m_j),
where d = 2 if PLANAR, d = 3 if SPATIAL

jVar is that j that realizes the above maximum
*/
{
    int dimM = bodies.N - 1;
#ifdef PLANAR
    MyDouble dpmiN[dimM][3];
#endif
#ifdef SPATIAL
    MyDouble dpmiN[dimM][6];
#endif

    // auxiliary matrix to remember constant parts for i:
    // dpmiN[i][0] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(x_i - x_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][1] = 3 \frac{(x_i - x_(N-1))(y_i - y_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][2] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(y_i - y_(N-1))^2}{r_{i(N-1)}^5}
    // #ifdef SPATIAL
    // dpmiN[i][3] = \frac{1}{r_{i(N-1)}^3} - 3\frac{(z_i - z_(N-1))^2}{r_{i(N-1)}^5}
    // dpmiN[i][4] = 2 \frac{(x_i - x_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}
    // dpmiN[i][5] = 3 \frac{(y_i - y_(N-1))(z_i - z_(N-1))}{r_{i(N-1)}^5}
    fillAuxMatrix(bodies, dpmiN);

    jVar = 0;
    MyDouble dfdmj = 0.0;
    MyDouble temp = 0.0;

    for (int j = 0; j < dimM; j++)
    // max_{j\in {0, ..., N-2}}
    {
#ifdef PLANAR
        temp = dfdmjPLANAR(bodies, j, dpmiN);
#endif
#ifdef SPATIAL
        temp = dfdmjSPATIAL(bodies, j, dpmiN); // |df/dm_j|*diam(m_j)
#endif

        if (dfdmj < temp)
        {
            dfdmj = temp;
            jVar = j;
        }
    }
    return dfdmj;
}

MyDouble fixDirection(const Bodies &bodies, MyMatrix &DFF, int size, int &varMax, int &coord)
// zwraca maksymalną wyliczona wartosc |dfdx_i|*diam(x_i)
//
// liczę normy kolumny macierzy DFF pomnozone przez srednice zmiennej
// varMax is a variable x_i where |dfdx_i|*diam(x_i) is maksimal
// coord points if it is x, y or z
{
    MyDouble dfdxNorm = 0.0;
    MyDouble dfdxMax = 0.0;
    int jVar = 0;
    int jCoord = 0;

    // na wszelki wypadek inicjuje na zero
    varMax = 0;
    coord = 0;
    MyInterval d;

    //    int N = bodies.N;

    for (int j = 0; j < size; j++)
    {
        vectorIndicesToNumberOfBodies(j, jVar, jCoord);
        switch (jCoord)
        {
        case 0:
            d = diam(bodies.x(jVar));
            break;

        case 1:
            d = diam(bodies.y(jVar));
            break;

#ifdef SPATIAL
        case 2:
            d = diam(bodies.z(jVar));
            break;
#endif

        default:
            break;
        }

        dfdxNorm = midPointInt(dfdxNormL1(DFF, j, size) * d);
        // ||df/dx_j||*diam(x_j);

        if (dfdxNorm > dfdxMax)
        {
            varMax = jVar;
            coord = jCoord;
            dfdxMax = dfdxNorm;
        }
    }

#ifdef MASSDIV
    MyDouble maxdxdMj = dxdMjNorm(bodies, jVar);
    if (maxdxdMj > dfdxMax)
    {
        varMax = jVar;
        coord = 3;
        dfdxMax = maxdxdMj;
    }
#endif

    return dfdxMax;
}
